@@ -11,6 +11,9 @@
 #include "scalarapi.c"
 
 #include "datetime.c"
+#include "datetime_strings.c"
+#include "datetime_busday.c"
+#include "datetime_busdaycal.c"
 #include "arraytypes.c"
 
 #include "hashdescr.c"
@@ -18,6 +21,7 @@
 
 #include "descriptor.c"
 #include "flagsobject.c"
+#include "alloc.c"
 #include "ctors.c"
 #include "iterators.c"
 #include "mapping.c"
@@ -35,11 +39,20 @@
 #include "conversion_utils.c"
 #include "buffer.c"
 
-
-#ifndef Py_UNICODE_WIDE
+#include "nditer_constr.c"
+#include "nditer_api.c"
+#include "nditer_templ.c"
+#include "nditer_pywrap.c"
+#include "lowlevel_strided_loops.c"
+#include "dtype_transfer.c"
+#include "einsum.c"
+#include "array_assign.c"
+#include "array_assign_scalar.c"
+#include "array_assign_array.c"
 #include "ucsnarrow.c"
-#endif
 
 #include "arrayobject.c"
 
+#include "numpymemoryview.c"
+
 #include "multiarraymodule.c"
